/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2004 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * Portions of this software are based upon public domain software
 * originally written at the National Center for Supercomputing Applications,
 * University of Illinois, Urbana-Champaign.
 */

/**
 * @mainpage XHTML Negotiation Module
 * mod_xhtml_neg -- a module to negotiate an appropriate content-type for
 * XHTML content. Clients that request application/xhtml+xml will be returned
 * it, other clients will receive text/html for backward compatibility.
 * Other content types, such as application/xml can be returned to suitable
 * clients.
 *
 * Some special cases for Internet Explorer compatibility were inevitably
 * introduced. See XhtmlNegStarsIgnore directive.
 *
 * Specs followed include RFC 2616, RFC 3023, RFC 3236.
 * Also deals with issues contained in RFC 2854.
 *
 * From http://www.w3.org/2003/01/xhtml-mimetype/content-negotiation
 * Etags should be unique for each content-type returned. This is done by
 * adding our own suffix to Etag headers when a negotiation is successful.
 * The suffix is a hash of the content-type value returned to the client.
 *
 * @author Nicholas Cull <run2000@users.sourceforge.net>
 * @date 2 January 2008
 * @version 1.0a
 *
 * @history
 * 0.5 Basic handling of Accept header \n
 * 0.6 Added charset handling \n
 * 0.7 Added profile handling, fixed up default charset for text/xml \n
 * 0.8 First public release. \n
 * 0.9 ETag handled correctly for different content types. \n
 * 0.92 Pick up the AddDefaultCharset directive from the Apache core \n
 * 0.93 Use a better hash function for ETag uniqueness \n
 * 0.94 Parse q-values correctly irrespective of system locale \n
 * 1.0 pre 1 Prepare for stable release \n
 * 1.0 First stable release \n
 * 1.0a Small fix to avoid processing subrequests (Apache 2.x only) \n
 *
 * @directive
 * @c XhtmlNegActive Is this module enabled? Defaults to no. \n
 * @c XhtmlNegLog    A file name for an output log file \n
 * @c XhtmlNegTypes  A file extension followed by one or more matching
 *                   content-type strings \n
 * @c XhtmlNegStarsIgnore The number of stars in an Accept token which should
 *                        be ignored if we match them \n
 * @c XhtmlNegCache Should negotiated HTTP 1.0 requests be cacheable?
 *                  Defaults to no \n
 */

/**
 * @file
 * The main source file for the mod_xhtml_neg module. This file implements
 * the module for Apache 2.x servers.
 */

#include "apr.h"
#include "apr_strings.h"
#include "apr_lib.h"

#define APR_WANT_STRFUNC
#include "apr_want.h"

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_request.h"

/* This is naughty, because it gives us access to private areas of
 * http_core.c. We use this to retrieve default charset information
 * from the core module. This is populated by the AddDefaultCharset
 * directive.
 */
#define CORE_PRIVATE

#include "http_core.h"
#include "http_protocol.h"
#include "http_log.h"
#include "lookupa.h"

/** Read/write flags used when opening the log file. */
static int xfer_flags = (APR_WRITE | APR_APPEND | APR_CREATE);

/** Permission flags used when opening the log file. */
static apr_fileperms_t xfer_perms = APR_OS_DEFAULT;

/* #define MOD_XHTML_NEG_DEBUG 1 */

/** The default HTTP character set for most content types. */
#define DEFAULT_CHARSET_NAME "iso-8859-1"

/** The default HTTP character set for text/xml and text/xml-external-entity
  * content types. */
#define DEFAULT_TEXT_XML_CHARSET "us-ascii"

/**
 * Test whether a string is empty without having to do a full scan of the
 * string just to return a length, as per strlen.
 *
 * @param str a string pointer, possibly NULL, to be tested
 * @return a true value if the pointer is NULL, or points to a zero-length
 * string, otherwise a false value
 */
#define mod_xhtml_strempty(str) (((str) == NULL) || (*(str) == '\0'))

/**
 * Enumeration for determining whether processing should be enabled for
 * this module. By default, we set the directory config state to
 * active_dontcare, meaning the value hasn't been explicitly set. When
 * it is set by XhtmlNegActive, the directory config will be set to either
 * active_on or active_off.
 */
typedef enum {
    active_on = 1,          /**< processing should be active */
    active_off = 0,         /**< processing should not be active */
    active_dontcare = 2     /**< default processing state (not active) */
} xhtml_neg_active;

/**
 * Enumeration for determining whether HTTP 1.0 caching should be allowed
 * for negotiated documents. By default, we set the server config state to
 * caching_dontcare, meaning the value hasn't been explicitly set. When it
 * set by the XhtmlNegCache parameter, the server config will be set to
 * either caching_on or caching_off.
 */
typedef enum {
    caching_on = 1,         /**< HTTP 1.0 caching should be active */
    caching_off = 0,        /**< HTTP 1.0 caching should not be active */
    caching_dontcare = 2    /**< default processing state (not active) */
} http_caching;


module AP_MODULE_DECLARE_DATA xhtml_neg_module;

/**
 * A per-server configuration state for this module. Here we record any
 * log file that we want to write to, the open file descriptor, and the
 * caching state for HTTP 1.0 requests.
 */
typedef struct {
    char *fname;          /**< Name of the log file we write to, if any */
    apr_file_t *log_fd;   /**< File descriptor of the log file */
    http_caching cache;   /**< Flag to cache negotiated content in
                            *  HTTP 1.0 requests */
} xhtml_neg_state;

/**
 * A per-directory configuration state for this module. Here we record the
 * file extension to content-type mappings, whether the module should be
 * active for this directory, and the types of wildcard Accept tokens that
 * we should ignore.
 */
typedef struct {
    xhtml_neg_active active;  /**< Is the module active for this directory? */
    apr_array_header_t *extensions; /**< An array of extension_rec records */
    int stars_ignore;         /**< How many content-type stars before
                                *  we ignore */
} xhtml_dir_config;

/**
 * Record of available info on a media type specified by the client.
 * We also use it for encoding and profile information.
 */
typedef struct {
    char *name;             /**< The name, normalised to lowercase */
    float q_value;          /**< Quality value, from 0 to 1. */
    char *charset;          /**< Charset parameter, for content-type only */
    char *profile;          /**< The XHTML profile, as per RFC 3236 */
    int stars;              /**< How many stars in this accept token? */
    char *hashcode;         /**< The hashcode to be appended to Etags */
} accept_rec;

/**
 * Record the maps a file extension to zero or more content-types. The content
 * types themselves are stored as an array of accept_rec records.
 */
typedef struct {
    char *extension;                   /**< Filename extension */
    apr_array_header_t *content_types; /**< Array of content accept tokens */
} extension_rec;

/*
 * Utility functions for strings, since we can't rely on these functions being
 * included.
 */

/**
 * Calculate the length of the given string. The pointer may be NULL.
 *
 * @param str a string pointer, possibly NULL, to be tested
 * @return the length of the string if the pointer is non-NULL, otherwise
 * zero
 */

static int mod_xhtml_strlen( const char *str ) {
   int len = 0;

   if( str == NULL ) {
       return len;
   }

   while( str[len] != '\0' ) {
      len++;
   }

   return len;
}

/**
 * Compare one string against another in a case-sensitive manner.
 * Will correctly deal with either pointer being NULL.
 *
 * @param str1 the first string pointer to be compared
 * @param str2 the second string pointer to be compared
 * @return zero if the strings are identical, less than zero if the
 * first string is less than the second, or greater than zero if the
 * first string is greater than the second
 */

static int mod_xhtml_strcmp( const char *str1, const char *str2 ) {
   int offset = 0;
   char ch1, ch2;
   int diff;

   /* Special case for either string being NULL. */
   if( str1 == NULL ) {
      if( str2 == NULL ) {
          return 0;
      }
      return -1;
   } else if( str2 == NULL ) {
      return 1;
   }

   while( 1 ) {
      ch1 = str1[ offset ];
      ch2 = str2[ offset ];
      diff = ch2 - ch1;
      if( diff ) {
         return diff;
      }
      /* Return if we encounter end-of-string. */
      if( ch1 == '\0' ) {
         return 0;
      }
      offset++;
   }
}

/**
 * Compare one string against another in a case-insensitive manner.
 * Will correctly deal with either pointer being NULL.
 *
 * @param str1 the first string pointer to be compared
 * @param str2 the second string pointer to be compared
 * @return zero if the strings are identical, less than zero if the
 * first string is less than the second, or greater than zero if the
 * first string is greater than the second
 */

static int mod_xhtml_stricmp( const char *str1, const char *str2 ) {
   int offset = 0;
   char ch1, ch2;
   int diff;

   /* Special case for either string being NULL. */
   if( str1 == NULL ) {
      if( str2 == NULL ) {
          return 0;
      }
      return -1;
   } else if( str2 == NULL ) {
      return 1;
   }

   while( 1 ) {
      ch1 = str1[ offset ];
      ch2 = str2[ offset ];

      /* Set both characters to lowercase (ASCII) if necessary. */
      if(( ch1 >= 65 ) && ( ch1 <= 90 )) {
         ch1 += 32;
      }
      if(( ch2 >= 65 ) && ( ch2 <= 90 )) {
         ch2 += 32;
      }

      diff = ch2 - ch1;
      if( diff ) {
         return diff;
      }
      /* Return if we encounter end-of-string. */
      if( ch1 == '\0' ) {
         return 0;
      }
      offset++;
   }
}

/**
 * Compare strings up to a specified limit in a case-sensitive manner.
 * If a NULL byte is encountered, finish comparing at that character.
 * Will correctly deal with either pointer being NULL.
 *
 * @param str1 the first string pointer to be compared
 * @param str2 the second string pointer to be compared
 * @param len the maximum number of characters to be compared
 * @return zero if the substrings are identical, less than zero if the
 * first substring is less than the second, or greater than zero if the
 * first substring is greater than the second
 */

static int mod_xhtml_strncmp( const char *str1, const char *str2,
                              const int len ) {
   int offset = 0;
   char ch1, ch2;
   int diff;

   /* Special case for either string being NULL. */
   if( str1 == NULL ) {
      if( str2 == NULL ) {
          return 0;
      }
      return -1;
   } else if( str2 == NULL ) {
      return 1;
   }

   while( offset < len ) {
      ch1 = str1[ offset ];
      ch2 = str2[ offset ];
      diff = ch2 - ch1;
      if( diff ) {
         return diff;
      }
      /* Allow early return if we encounter end-of-string. */
      if( ch1 == '\0' ) {
         return 0;
      }
      offset++;
   }
   return 0;
}

/**
 * Compare strings up to a specified limit in a case-insensitive manner.
 * If a NULL byte is encountered, finish comparing at that character.
 * Will correctly deal with either pointer being NULL.
 *
 * @param str1 the first string pointer to be compared
 * @param str2 the second string pointer to be compared
 * @param len the maximum number of characters to be compared
 * @return zero if the substrings are identical, less than zero if the
 * first substring is less than the second, or greater than zero if the
 * first substring is greater than the second
 */

static int mod_xhtml_strnicmp( const char *str1, const char *str2,
                               const int len ) {
   int offset = 0;
   char ch1, ch2;
   int diff;

   /* Special case for either string being NULL. */
   if( str1 == NULL ) {
      if( str2 == NULL ) {
          return 0;
      }
      return -1;
   } else if( str2 == NULL ) {
      return 1;
   }

   while( offset < len ) {
      ch1 = str1[ offset ];
      ch2 = str2[ offset ];

      /* Set both characters to lowercase (ASCII) if necessary. */
      if(( ch1 >= 65 ) && ( ch1 <= 90 )) {
         ch1 += 32;
      }
      if(( ch2 >= 65 ) && ( ch2 <= 90 )) {
         ch2 += 32;
      }

      diff = ch2 - ch1;
      if( diff ) {
         return diff;
      }
      /* Allow early return if we encounter end-of-string. */
      if( ch1 == '\0' ) {
         return 0;
      }
      offset++;
   }
   return 0;

}

/**
 * Determines if the given string end with the given suffix. Case matching
 * can be enabled or disabled.
 *
 * @param reference the reference string to be tested, possibly NULL
 * @param suffix the suffix to test against, possibly NULL
 * @param casecompare non-zero if a case sensitive compare is wanted,
 * otherwise zero for a case insensitive compare
 * @return non-zero if the suffix matches, otherwise zero
 */

static int mod_xhtml_strendswith( const char *reference, const char *suffix,
                        const int casecompare ) {
    int ref_length;
    int suffix_length;
    const char *startpos;

    if(( reference == NULL ) || ( suffix == NULL )) {
        return 0;
    }

    ref_length = mod_xhtml_strlen( reference );
    suffix_length = mod_xhtml_strlen( suffix );

    /** Is the reference string at least as long as the suffix? */
    if( ref_length < suffix_length ) {
        return 0;
    }

    startpos = reference + ( ref_length - suffix_length );

    if( casecompare ) {
        if( mod_xhtml_strncmp( startpos, suffix, suffix_length ) == 0 ) {
            return 1;
        }
    } else {
        if( mod_xhtml_strnicmp( startpos, suffix, suffix_length ) == 0 ) {
            return 1;
        }
    }

    return 0;
}

/*
 * Taken from a mod_negotiation fix on the Apache 2.1 branch.
 */

/**
 * Parse quality value. atof(3) is not globally usable here, because it
 * depends on the locale (argh).
 *
 * However, RFC 2616 states: \n
 * 3.9 Quality Values
 *
 * [...] HTTP/1.1 applications MUST NOT generate more than three digits
 * after the decimal point. User configuration of these values SHOULD also
 * be limited in this fashion.
 *
 *     qvalue         = ( "0" [ "." 0*3DIGIT ] )
 *                    | ( "1" [ "." 0*3("0") ] )
 *
 * This is quite easy. If the supplied string doesn't match the above
 * definition (loosely), we simply return 1 (same as if there's no q-value)
 *
 * @param string a string containing a q-value to be parsed
 * @return the parsed q-value, truncated to three decimal places if necessary
 */

static float mod_xhtml_atoq(const char *string)
{
    if (mod_xhtml_strempty(string)) {
        return  1.0f;
    }

    while (*string && apr_isspace(*string)) {
        ++string;
    }

    /* Be tolerant and accept q-values without leading zero
     * (also for backwards compatibility, where atof() was in use)
     * This also handles the 1.0 case.
     */
    if (*string != '.' && *string++ != '0') {
        return 1.0f;
    }

    if (*string == '.') {
        /* Better only one division later, than dealing with fscking
         * IEEE format 0.1 factors ...
         */
        int i = 0;

        if (*++string >= '0' && *string <= '9') {
            i += (*string - '0') * 100;

            if (*++string >= '0' && *string <= '9') {
                i += (*string - '0') * 10;

                if (*++string > '0' && *string <= '9') {
                    i += (*string - '0');
                }
            }
        }

        return (float)i / 1000.0f;
    }

    return 0.0f;
}

/**
 * This naughty function goes digging into the Apache http_core module
 * to find the default character set.
 *
 * @param r the current HTTP request from which we can determine the
 * configuration of the Apache core
 * @return the default character set as configured in the Apache core,
 * or NULL if no value is configured
 * @todo Is there a cleaner way to find the default charset information?
 */

static const char *get_default_charset( request_rec *r )
{
    core_dir_config *conf;

    conf = (core_dir_config *)
        ap_get_module_config(r->per_dir_config, &core_module);

    if( conf->add_default_charset == ADD_DEFAULT_CHARSET_ON ) {
        return conf->add_default_charset_name;
    }

    return NULL;
}

/**
 * Given an array of extension_rec records, find one with the given file
 * extension. If a match is found, return the extension_rec, otherwise
 * return NULL.
 *
 * @param extensions an array of extension_rec elements to be scanned
 * @param ext the file extension to match
 * @return the matching extension_rec item if one was found, otherwise NULL
 */

static extension_rec *find_extension( apr_array_header_t *extensions,
                                      const char *ext )
{
    extension_rec *list, *item;
    int i;

    if(( extensions == NULL ) || ( extensions->nelts == 0 )) {
        return NULL;
    }

    /*
     * Relying on locality of reference, an extension is most likely to be
     * the last item in the array. On that basis, scan backwards.
     */
    list = (extension_rec*) extensions->elts;
    for( i = extensions->nelts - 1; i >= 0; --i ) {
        item = &list[ i ];
        if( mod_xhtml_stricmp( item->extension, ext ) == 0 ) {
            return item;
        }
    }
    return NULL;
}

/**
 * Count the number of stars in an Accept content token. This helps determine
 * priority in case of a tie in q-value priorities.
 *
 * @param content_type the content token that we're interested in, possibly
 * NULL
 * @return the number of '*' characters in the string, or 0 if the string is
 * NULL
 */

static int count_stars(const char *content_type) {
    int stars = 0;
    int offset = 0;

    if( content_type == NULL ) {
        return 0;
    }

    while( content_type[ offset ] != '\0' ) {
        if( content_type[ offset ] == '*' ) {
            stars++;
        }
        offset++;
    }
    return stars;
}

/**
 * Reconstruct the original content-type, if required due to special "charset"
 * values. Normally the content-type will be a plain string without charset
 * encoding. In these cases, just return the content-type without modification.
 * In the case where we also have a charset entry, have to create the full
 * content-type string.
 *
 * @param p a memory pool from which we can allocate temporary memory for this
 * request
 * @param content_rec the accept_rec that matched the content negotiation
 * @return a string that contains a correctly formatted content-type value,
 * or NULL if there is no valid record
 * @todo should we send the "profile" parameter as well?
 */

static char *reconstruct_content_type( apr_pool_t *p, accept_rec *content_rec )
{
    if(( content_rec == NULL ) || ( content_rec->name == NULL )) {
        return NULL;
    }

    /* If we have a content-type, but no charset, return the original
     * content-type string. */
    if( mod_xhtml_strempty( content_rec->charset )) {
        return content_rec->name;
    }

    /* If we have a charset and a content-type, recreate the full string. */
    return apr_pstrcat( p, content_rec->name, "; charset=",
                       content_rec->charset, NULL );
}

/**
 * Make a simple hash code for the given accept_rec structure. This allows
 * us to generate unique Etags for accepted content-types. The Etag
 * consists of a hash of each character. This is then represented as a
 * zero padded hexadecimal number.
 *
 * The data that we use to generate the hash is the content-type and charset,
 * concatenated as one long string.
 *
 * The hash function is a hash in the public domain by By Bob Jenkins. See
 * lookupa.c or http://burtleburtle.net/bob/hash/doobs.html for details.
 *
 * @note Since we're only using this algorithm for Etag uniqueness rather
 * than for a hashtable lookup, DoS attacks such as those described in
 * http://www.cs.rice.edu/~scrosby/hash/ are not an issue here, especially
 * since our hash keys come from data in either httpd.conf or .htaccess files.
 *
 * @param p a memory pool from which we can allocate temporary memory for this
 * request
 * @param rec the accept_rec that matched the content negotiation; the hash
 * value will be added to this structure
 * @todo If the content-type returned ever includes the "profile" parameter,
 * include it as part of the hash code.
 */

static void make_etag_hashcode( apr_pool_t *p, accept_rec *rec )
{
    char *data;
    unsigned long hash;
    int len;

    data = apr_pstrcat( p, rec->name, ";", rec->charset, NULL );
    len = mod_xhtml_strlen( data );
    hash = lookup( data, len, 0 ) & 0xffffffff;

    rec->hashcode = apr_psprintf(p, "\"%08lx\"", (unsigned long) hash);
}

/**
 * Make an accept_rec with an explicit charset. If the record already
 * contains a charset parameter, it is returned as-is. Otherwise, the
 * parameter default_charset is used.
 *
 * @note default_charset can be NULL, in which case charset will be set to
 * the empty string.
 *
 * @param p a memory pool from which we can allocate temporary memory for this
 * request
 * @param rec the accept_rec that matched the content negotiation
 * @param default_charset the default character set as configured in the
 * Apache core
 * @return the original accept_rec if it already contains a charset parameter,
 * otherwise a new accept_rec containing the same information and the default
 * charset parameter
 */

static accept_rec *make_default_accept( apr_pool_t *p, accept_rec *rec,
                                        const char *default_charset )
{
     accept_rec *newrec;

     if(( rec == NULL ) || ( ! mod_xhtml_strempty( rec->charset ))) {
         return rec;
     }

     newrec = (accept_rec *) apr_palloc(p, sizeof (accept_rec));
     newrec->name = rec->name;
     newrec->q_value = rec->q_value;
     newrec->charset = ( default_charset ?
                         apr_pstrdup( p, default_charset ) : "" );
     newrec->profile = rec->profile;
     newrec->stars = rec->stars;
     make_etag_hashcode( p, newrec );

     return newrec;
}

/**
 * Merge vlist_validators from different modules. This code is adapted
 * from code in http_protocol.c in the Apache 1.3 core.
 *
 * @param p a memory pool from which we can allocate temporary memory for this
 * request
 * @param old_variant an Etag variant taken from the vlist_validator parameter
 * of the current request
 * @param new_variant an Etag variant as supplied by make_etag_hashcode
 * @return a correctly merged Etag variant, taking into account whether either
 * variant has a weak validator attached, or whether either variant is NULL
 */

static char *merge_validators(apr_pool_t *p, char *old_variant,
                                             char *new_variant)
{
    int old_weak, new_weak;

    if (mod_xhtml_strempty(old_variant)) {
        return new_variant;
    } else if (mod_xhtml_strempty(new_variant)) {
        return old_variant;
    }

    /* Merge any two vlist entries: one from any previous module, and
     * one from the current module. This merging makes revalidation
     * somewhat safer, ensures that caches which can deal with
     * Vary will (eventually) be updated if the set of variants is
     * changed, and is also a protocol requirement for transparent
     * content negotiation.
     */

    /* if the variant list validator is weak, we make the whole
     * structured etag weak.  If we would not, then clients could
     * have problems merging range responses if we have different
     * variants with the same non-globally-unique strong etag.
     */
    old_weak = (old_variant[0] == 'W');
    new_weak = (new_variant[0] == 'W');

    /* merge old and new variant_etags into a structured etag */
    old_variant[mod_xhtml_strlen(old_variant) - 1] = '\0';
    if ((new_weak) && (mod_xhtml_strlen(new_variant) > 3))
        new_variant += 3;
    else
        new_variant++;

    if((old_weak == 0) && (new_weak != 0)) {
        return apr_pstrcat(p, "W/", old_variant, ";", new_variant, NULL);
    }

    return apr_pstrcat(p, old_variant, ";", new_variant, NULL);
}


/**
 * For a given filename, scan through an array of extension_rec records
 * until we find a match. The first match wins. If no match is found,
 * NULL is returned.
 *
 * @param extensions an array of extension_rec elements containing filename
 * suffixes to be matched
 * @param filename the filename we want to match
 * @return the matching extension_rec item if one exists, otherwise NULL
 */

static extension_rec *get_extension_for_filename(
        apr_array_header_t *extensions, const char *filename )
{
    extension_rec *list, *item;
    int i;

    list = (extension_rec *)extensions->elts;
    for(i = 0; i < extensions->nelts; i++) {
        item = &list[ i ];
        if( mod_xhtml_strendswith( filename, item->extension, 0 )) {
            return item;
        }
    }
    return NULL;
}

/* Following code modified from mod_negotiate.c */

/**
 * Get a single mime type entry --- one media type and parameters;
 * enter the values we recognize into the argument accept_rec.
 *
 * @param p a memory pool from which we can allocate temporary memory for this
 * request
 * @param result the accept_rec to be populated by parsing the accept_line
 * @param accept_line a string containing the accept token to be parsed
 * @return any remaining accept_line string left to be parsed
 */

static const char *get_entry(apr_pool_t *p, accept_rec *result,
                             const char *accept_line)
{
    result->q_value = 1.0f;
    result->charset = "";
    result->profile = "";
    result->stars = 0;
    result->hashcode = "";

    /*
     * Note that this handles what I gather is the "old format",
     *
     *    Accept: text/html text/plain moo/zot
     *
     * without any compatibility kludges --- if the token after the
     * MIME type begins with a semicolon, we know we're looking at parms,
     * otherwise, we know we aren't.  (So why all the pissing and moaning
     * in the CERN server code?  I must be missing something).
     */

    result->name = ap_get_token(p, &accept_line, 0);
    ap_str_tolower(result->name);     /* You want case-insensitive,
                                       * you'll *get* case-insensitive.
                                       */

    while (*accept_line == ';') {
        /* Parameters ... */

        char *parm;
        char *cp;
        char *end;

        ++accept_line;
        parm = ap_get_token(p, &accept_line, 1);

        /* Look for 'var = value' --- and make sure the var is in lcase. */

        for (cp = parm; (*cp && !apr_isspace(*cp) && *cp != '='); ++cp) {
            *cp = apr_tolower(*cp);
        }

        if (!*cp) {
            continue;           /* No '='; just ignore it. */
        }

        *cp++ = '\0';           /* Delimit var */
        while (*cp && (apr_isspace(*cp) || *cp == '=')) {
            ++cp;
        }

        if (*cp == '"') {
            ++cp;
            for (end = cp;
                 (*end && *end != '\n' && *end != '\r' && *end != '\"');
                 end++);
        } else {
            for (end = cp; (*end && !apr_isspace(*end)); end++);
        }
        if (*end) {
            *end = '\0';        /* strip ending quote or return */
        }
        ap_str_tolower(cp);

        if (parm[0] == 'q'
            && (parm[1] == '\0' || (parm[1] == 's' && parm[2] == '\0'))) {
            result->q_value = (float)mod_xhtml_atoq(cp);
        }
        else if (mod_xhtml_strcmp(parm, "charset") == 0) {
            result->charset = cp;
        }
        else if (mod_xhtml_strcmp(parm, "profile") == 0) {
            result->profile = cp;
        }
    }

    if (*accept_line == ',') {
        ++accept_line;
    }

    /* Count stars to determine how specific the match is. */
    result->stars = count_stars( result->name );

    return accept_line;
}

/**
 * Parse and tokenise Accept header lines.
 *
 * The Accept request header is handled by do_accept_line() - it has the
 * basic structure of a list of items of the format:
 *
 *    name; q=N; charset=TEXT; profile="URI"
 *
 * Since this header is very similar in structure to the Accept-Charset
 * request header, we can reuse the parsing in get_entry.
 *
 * The profile parameter was added to handle section 8 of RFC 3236.
 * These probably won't appear at an origin server, but we handle
 * them explicitly if they do.
 *
 * @param p a memory pool from which we can allocate temporary memory for this
 * request
 * @param accept_line the HTTP Accept request value to be parsed
 * @return an array of accept_rec items containing the parsed Accept tokens
 */

static apr_array_header_t *do_accept_line(apr_pool_t *p,
                                          const char *accept_line)
{
    apr_array_header_t *accept_recs;

    if (accept_line == NULL) {
        return NULL;
    }

    accept_recs = apr_array_make(p, 40, sizeof(accept_rec));

    while (*accept_line) {
        accept_rec *new = (accept_rec *) apr_array_push(accept_recs);
        accept_line = get_entry(p, new, accept_line);
    }

    return accept_recs;
}

/**
 * Parse and tokenise Accept-Charset headers. There is an extra record
 * inserted when neither "*" nor "ISO-8859-1" are mentioned.
 *
 * @param p a memory pool from which we can allocate temporary memory for this
 * request
 * @param accept_charset_line the HTTP Accept-Charset request value to be
 * parsed
 * @return an array of accept_rec items containing the parsed Accept-Charset
 * tokens
 */

static apr_array_header_t *do_accept_charset_line(apr_pool_t *p,
        const char *accept_charset_line)
{
    apr_array_header_t *accept_recs;
    int star_found, iso8859_found;

    if (accept_charset_line == NULL) {
        return NULL;
    }

    accept_recs = apr_array_make(p, 40, sizeof(accept_rec));
    star_found = 0;
    iso8859_found = 0;

    /* Scan all the accept_recs tokens for a name of "ISO-8859-1" (case
     * insensitive). If there are no matching records, construct one with
     * a q value of "1", as per section 14.2 of the HTTP 1.1 spec (RFC 2616).
     * Also have to take into account whether "*" is present (this point is
     * not so clear in the spec).
     */
    while (*accept_charset_line) {
        accept_rec *new = (accept_rec *) apr_array_push(accept_recs);
        accept_charset_line = get_entry(p, new, accept_charset_line);

        if( new->stars > 0 ) {
            star_found++;
        } else if( mod_xhtml_stricmp( new->name, DEFAULT_CHARSET_NAME )
                == 0 ) {
            iso8859_found++;
        }
    }

    /** Merge new file extensions into new array */
    if(( star_found == 0 ) && ( iso8859_found == 0 )) {
        /* Add new default iso-8859-1 entry to the array with quality of 1. */
        accept_rec *new = (accept_rec *) apr_array_push(accept_recs);
        new->name = DEFAULT_CHARSET_NAME;
        new->q_value = 1.0f;
        new->charset = "";
        new->profile = "";
        new->stars = 0;
        new->hashcode = "";
    }

    return accept_recs;
}

/**
 * Look for a matching charset within the given array of accept_recs. If we
 * find a match, return the corresponding q value, otherwise return 0.0.
 *
 * @param accept_charsets an array of accept_rec items containing
 * Accept-Charset tokens to be matched
 * @param content_charset the target charset parameter to match
 * @return the q value of the best match, or 0.0f if no match was found
 */

static float charsets_match( apr_array_header_t *accept_charsets,
                             const char *content_charset )
{
    accept_rec *list, *item;
    int i;
    float default_q = 0.0f;

    /* An empty Accept-Charset header means the client accepts anything. */
    if(( accept_charsets == NULL ) || ( accept_charsets->nelts == 0 )) {
        return 1.0f;
    }

    list = (accept_rec*)accept_charsets->elts;
    for (i = 0; i < accept_charsets->nelts; ++i) {
        item = &list[i];
        if( item->stars > 0 ) {
            default_q = item->q_value;
        } else if( mod_xhtml_stricmp( item->name, content_charset ) == 0 ) {
            return item->q_value;
        }
    }
    return default_q;
}

/**
 * Determine if a given content_type record matches the criteria of the given
 * content_accept record. Returns a quality value for the given variables.
 *
 * First, check whether names match, depending on whether any stars are
 * present in the content_accept record. If the first character is a star,
 * then we assume its a total wildcard match. Otherwise, look for the major
 * content-type only.
 *
 * Once we match based on names, check whether a charset is present in the
 * content_accept record. If so, need to match charsets exactly. Otherwise,
 * no further check is needed, and we can return true immediately.
 *
 * The quality value returned is a straight multiplication of all the q-values
 * the went into determining the result: the q value from the config file
 * (if any was specified), the q value of the Accept content-type header, and
 * the q value of the Accept-Charset header. Where no q-value is specified,
 * the value "1.0" is used.
 *
 * @param content_type the content-type token to match
 * @param content_accept the accept token to be matched
 * @param accept_charsets an array of accept_rec items containing
 * Accept-Charset tokens to be matched
 * @param default_charset the default charset value as configured in the
 * Apache core
 * @return the q value of the best matching content-type and charset
 * combination, or 0.0f if no match was found
 */

static float types_match( accept_rec *content_type,
                        accept_rec *content_accept,
                        apr_array_header_t *accept_charsets,
                        const char *default_charset )
{
    int offset = 0;
    const char *content_charset;
    char *accept_profile, *content_profile;
    float charset_q, accept_q, server_q;

    if( content_accept->stars == 0 ) {
        /* Match based on exact name match
          (remember, both lowercase by now). */
       if( mod_xhtml_strcmp( content_accept->name, content_type->name )
               != 0 ) {
           return 0.0f;
       }
    } else {
        if( content_accept->name[0] == '*' ) {
            /* Assume total wildcard match, so no need to look further. */
        } else {
            /* Have to do match based on major content-type only. */
            while( content_accept->name[offset] != '\0' ) {
               if ( content_accept->name[offset++] == '/' ) {
                   break;
               }
            }
            if( mod_xhtml_strncmp( content_accept->name,
                          content_type->name, offset ) != 0 ) {
                return 0.0f;
            }
        }
    }

    /* We found a matching content-type. Now compare profiles and charsets. */
    accept_q = content_accept->q_value;
    server_q = content_type->q_value;

    /* Check for "profile" optional parameter. */
    content_profile = content_type->profile;
    accept_profile = content_accept->profile;

    if( mod_xhtml_strempty( accept_profile )) {
        /* Accept profile is specified in the request. */
        if( ! mod_xhtml_strempty( content_profile )) {
            /* Accept token specifies a profile, but server does not.
             * Have to reject. (RFC 3236 doesn't specify a default.)
             *
             * Note that although origin servers don't have to specify
             * a profile, the solution would be to specify the particular
             * profile with a q value of 1.0, and a generic XHTML
             * response without a profile and a (possibly) lower q value.
             * This can happen at either the client or server end.
             */
            return 0.0f;
        }
    } else if( ! mod_xhtml_strempty( content_profile )) {
        /* Both client and server specify profiles. Look for match.
         * Note that we have to use case-sensitive matching, since the
         * profile string is typically a URL.
         */
        if( mod_xhtml_strcmp( content_profile, accept_profile ) != 0 ) {
            /* Profile parameters are different. Reject. */
            return 0.0f;
        }
    }

    /* Step 1 is to set a default content-type for the response. If none
     * was specified in the content-type, set it to "iso-8859-1" or
     * "US-ASCII".
     */
    if( mod_xhtml_strempty( content_type->charset )) {
        /* Sigh... RFC 3023 comes into play here. For most content-types
         * RFC 2616 specifies the default character set should be
         * ISO-8859-1 unless specifically overridden by the charset parameter.
         * For the content-types "text/xml" and
         * "text/xml-external-parsed-entity" section 3.6 of RFC 3023
         * say US-ASCII must be assumed, and explicitly overrides RFC 2616.
         * See also example 8.5 of RFC 3023.
         */
        if( ! mod_xhtml_strempty( default_charset )) {
            content_charset = default_charset;
        } else if( mod_xhtml_strnicmp( content_type->name, "text/xml", 8 )
                      == 0 ) {
            content_charset = DEFAULT_TEXT_XML_CHARSET;
        } else {
            content_charset = DEFAULT_CHARSET_NAME;
        }
    } else {
        content_charset = content_type->charset;
    }

    if( mod_xhtml_strempty( content_accept->charset )) {
        /* Accept token says we accept anything. Have to check for
         * Accept-Charset also.
         */
        charset_q = charsets_match( accept_charsets, content_charset );
        return accept_q * charset_q * server_q;

    } else if( mod_xhtml_stricmp( content_charset, content_accept->charset )
             == 0 ) {
        /* Accept "charset" parameter is non-empty, compare on given charset.
         * Note that the content_accept q value here covers both the
         * content-type and the charset value.
         */
        return accept_q * server_q;
    }

    return 0.0f;
}

/**
 * Implements the best match algorithm for a given Accept header, possible
 * content-type headers and an optional default content-type in the event of
 * a star-slash-star match.
 *
 * @param accept_type an array of accept_rec items containing Accept tokens
 * @param content_type an array of accept_rec items containing possible
 * content-type tokens to be matched
 * @param accept_charset an array of accept_rec items containing
 * Accept-Charset tokens
 * @param default_charset the default character set as configured in the
 * Apache core
 * @param stars_to_match the number of '*' tokens at which we start ignoring
 * the Accept token
 * @param xns the configuration state of the module in case we need to write
 * to the log file
 * @param p a memory pool from which we can allocate temporary memory for this
 * request
 * @return an accept_rec structure containing the best match we could find for
 * the current request, or NULL if we couldn't find a match
 */

static accept_rec *best_match(apr_array_header_t *accept_type,
                 apr_array_header_t *content_type,
                 apr_array_header_t *accept_charset,
                 const char *default_charset, int stars_to_match,
                 xhtml_neg_state *xns, apr_pool_t *p)
{
    float best_q = 0.0f;
    int best_stars = 10;
    int curr_stars;
    accept_rec *best_type = NULL;
    accept_rec *curr_accept, *curr_content;
    accept_rec *list_accept, *list_content;
    int i, j;
    float curr_q;
    char *logmessage;
    apr_size_t logsize;

    /* Deal with silly null pointers when Accept: header is empty */
    if((content_type == NULL) || (accept_type == NULL)) {
        return NULL;
    }

    /* Set up default star matching if none is specified in the
       config file. */
    if( stars_to_match < 0 ) {
        stars_to_match = 2;
    }

    /* This is a nasty, icky, order n times m algorithm. For each entry in the
     * content_type array, look through the accept_type array looking for
     * a match. If a match is found, we check best-q and best-stars to see if
     * the match is better than the one we have currently (higher q is better,
     * fewer stars are better). If the current match is superior, update
     * best_q, best_stars, and best_type records. In the case of a tie, the
     * existing record wins.
     *
     * This is also the place where we test for our limit on the number of
     * stars. We have to do it here because a star-slash-star match could
     * result in a high q value but fail the star limit test, while another
     * match could have a lower q value while passing the star limit test.
     *
     * At the end, if no suitable match was found, return the default type.
     * To the caller, this is indicated by a NULL result being returned.
     */
    list_content = (accept_rec*)content_type->elts;
    for (i = 0; i < content_type->nelts; ++i) {
        curr_content = &list_content[i];

        list_accept = (accept_rec*)accept_type->elts;
        for (j = 0; j < accept_type->nelts; ++j) {
            curr_accept = &list_accept[j];
            curr_stars = curr_accept->stars;
            curr_q = types_match( curr_content, curr_accept, accept_charset,
                                  default_charset );

#ifdef MOD_XHTML_NEG_DEBUG

            if(( curr_q > 0.0f ) && ( xns->log_fd != NULL )) {
                logmessage = apr_pstrcat(p,
                        "q_value found: ",
                        reconstruct_content_type(p, curr_content),
                        "\n", NULL);
                logsize = mod_xhtml_strlen( logmessage );

                apr_file_write(xns->log_fd, logmessage, &logsize);
            }
#endif

            if((( curr_q > best_q ) && ( curr_stars < stars_to_match )) ||
                  (( curr_q == best_q ) && ( curr_stars < best_stars ))) {
                best_q = curr_q;
                best_stars = curr_stars;
                best_type = curr_content;
            }
        }
    }

    /* If the best quality is 0, the client is rejecting our negotiated
     * content type, so return NULL.
     */
    if( best_q == 0.0f ) {
        best_type = NULL;
    }

    /* Return the reconstructed content-type record. */
    return make_default_accept( p, best_type, default_charset );
}

/**
 * For HTTP 1.1 responses, where the content has been modified based on HTTP
 * request headers, need to set the Vary header to name the headers that the
 * variation was based on. For some requests, this cannot be determined by
 * headers alone, and a "*" needs to be sent instead.
 *
 * Here we first check any existing Vary header for the presence of a "*".
 * If "Vary: *" exists, return it unaltered. Otherwise, we need to merge two
 * new tokens: "Accept" and "Accept-Charset".
 *
 * @param r the current HTTP request to which we merge Vary tokens
 */

static void set_vary_header(request_rec *r)
{
    apr_table_t *hdrs = r->headers_out;
    const char *vary = apr_table_get(hdrs, "Vary");

    if( vary == NULL )  {
        apr_table_setn( hdrs, "Vary", "Accept, Accept-Charset" );
        return;
    }

    /* Look for "Vary: *" header. */
    if( count_stars( vary ) > 0 ) {
        return;
    }

    apr_table_mergen( hdrs, "Vary", "Accept, Accept-Charset" );
}

/**
 * The main routine for the content-negotiation phase of this module goes
 * here. This is mainly setup, control flow and logging going on here.
 *
 * @note Due to the way we hook into the Apache handler system means
 * we always return DECLINED, even though handling was a success.
 * We do this so that the default handler will always run, and so that
 * ETag is handled correctly for both 200 OK and 304 Not Modified cases.
 *
 * @param r the current HTTP request
 * @return DECLINED, to indicate the request should be processed by the
 * next handler in the chain
 */

int xhtml_negotiate(request_rec *r)
{
    apr_array_header_t *extensions, *accept_types,
                       *accept_charsets, *content_types;
    const char *content_accept, *accept_charset, *default_charset;
    char *filename, *extension, *logmessage;
    apr_size_t logsize;
    xhtml_dir_config *conf;
    xhtml_neg_state *xns;
    extension_rec *item;
    apr_table_t *hdrs;
    accept_rec *result_rec;

    conf = (xhtml_dir_config *) ap_get_module_config(
            r->per_dir_config, &xhtml_neg_module);

    xns = ap_get_module_config (r->server->module_config, &xhtml_neg_module);

    if (conf == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                    "internal error -- no configuration: mod_xhtml_neg.c");
        return HTTP_INTERNAL_SERVER_ERROR;
    };

    /* Only work on the main request - no subrequests. */
    if( r->main != NULL ) {
        return DECLINED;
    }

    /* Return OK if either active isn't ON, or stars_ignore is 0. */
    if( conf->active != active_on ) {
        return DECLINED;
    }

    if( conf->stars_ignore == 0 ) {
        return DECLINED;
    }

    /* Only service requests for GET or HEAD methods. */
    if( r->method_number != M_GET ) {
        return DECLINED;
    }

    /* Retrieve the HTTP request headers table. */
    hdrs = r->headers_in;
    if( hdrs == NULL ) {
        return DECLINED;
    }

    /* If for some reason there's no filename, return. */
    filename = r->filename;
    if( mod_xhtml_strempty( filename )) {
        return DECLINED;
    }

    /* See if we can find the given file extension in the extensions array, */
    extensions = conf->extensions;
    extension = NULL;
    content_types = NULL;

    /* Find a matching extension for the given filename. */
    item = get_extension_for_filename( extensions, filename );
    if( item != NULL ) {
        extension = item->extension;
        content_types = item->content_types;
    } else {
       return DECLINED;
    }

    /* Do some logging... */
    if( xns->log_fd != NULL ) {

        logmessage = apr_pstrcat(r->pool,
                "Filename is ", filename, "\n",
                "File extension found -- ", extension, "\n", NULL );
        logsize = mod_xhtml_strlen( logmessage );
        apr_file_write(xns->log_fd, logmessage, &logsize );
    }

    /* Look up "accept" encoding header. */
    content_accept = apr_table_get(hdrs, "Accept");
    accept_charset = apr_table_get(hdrs, "Accept-Charset");
    default_charset = get_default_charset( r );

    /* More logging... */
    if( xns->log_fd != NULL ) {
        logmessage = apr_pstrcat(r->pool,
                "Content accept is: ",
                (content_accept ? content_accept : ""), "\n",
                "Accept charset is: ",
                (accept_charset ? accept_charset : ""), "\n",
                "Default charset is: ",
                (default_charset ? default_charset : ""), "\n",
                NULL);

        logsize = mod_xhtml_strlen( logmessage );
        apr_file_write(xns->log_fd, logmessage, &logsize );
    }

    /* Translate into content arrays. */
    accept_types = do_accept_line(r->pool, content_accept);
    accept_charsets = do_accept_charset_line(r->pool, accept_charset);

    /* Find accept_rec of best result, if any. */
    result_rec = best_match( accept_types, content_types,
                             accept_charsets, default_charset,
                             conf->stars_ignore, xns, r->pool );

    if( result_rec != NULL ) {
        /* Construct a new ETag suffix, and make sure it gets
         * appended to the ETag that would normally be returned
         * by Apache.
         */
        r->content_type = reconstruct_content_type( r->pool, result_rec );
        r->vlist_validator = merge_validators( r->pool, r->vlist_validator,
                                               result_rec->hashcode );

        if( xns->log_fd != NULL ) {
            logmessage = apr_pstrcat(r->pool,
                    "Best content-type: ", r->content_type, "\n",
                    "New validator is: ", r->vlist_validator, "\n",
                    NULL );
            logsize = mod_xhtml_strlen( logmessage );
            apr_file_write(xns->log_fd, logmessage, &logsize );
        }
    } else if( xns->log_fd != NULL ) {
        logmessage = "No matching content-type found, default used.\n";
        logsize = mod_xhtml_strlen( logmessage );
        apr_file_write(xns->log_fd, logmessage, &logsize);
    }

    /* Note that whether or not this module successfully identifies
     * a negotiated response, we should always send the appropriate
     * Vary or cache-control response.
     */
    if(r->proto_num < HTTP_VERSION(1,1)) {
        /* Really shouldn't cache responses for HTTP 1.0 or earlier.
         * Similarly to mod_negotiate, we make it configurable.
         */
        if( xns->cache != caching_on ) {
            r->no_cache = 1;
        }
    }
    set_vary_header(r);

    return DECLINED;
}


/* All Apache configuration file code goes here... */

/**
 * Is this module active for the given directory.
 *
 * @param in_dir_config the configuration information for this directory
 * @param active a yes/no flag indicating whether the module should be active
 * @return NULL to indicate that processing was successful
 */

static const char *set_xhtml_active(cmd_parms *cmd,
                                    void *in_dir_config, int active)
{
    xhtml_dir_config *dir_config = in_dir_config;

    /* If we're here at all it's because someone explicitly
     * set the active flag
     */
    if ( active ) {
        dir_config->active = active_on;
    } else {
        dir_config->active = active_off;
    }
    return NULL;
}

/**
 * Set whether HTTP 1.0 caching should be allowed. Default to "no" unless
 * specifically overridden.
 *
 * @param cmd the configuration information for this module on a per-server
 * basis
 * @param cache a flag indicating whether HTTP 1.0 caching should be enabled
 * @return NULL to indicate that processing was successful
 */

static const char *set_xhtml_cache_negotiated(cmd_parms *cmd, void *dummy,
                                              int cache)
{
    xhtml_neg_state *xns;

    xns = ap_get_module_config (cmd->server->module_config,
                                &xhtml_neg_module);

    if( cache ) {
        xns->cache = caching_on;
    } else {
        xns->cache = caching_off;
    }
    return NULL;
}

/**
 * Add content types for the given file extension.
 *
 * @param in_dir_config the configuration information for this directory
 * @param ext the file extension we're configuring
 * @param type the content type to be added for this file extension
 * @return NULL to indicate that processing was successful
 */

static const char *add_xhtml_type(cmd_parms *cmd, void *in_dir_config,
                                  const char *ext, const char *type)
{
    extension_rec *rec;
    accept_rec *new;
    xhtml_dir_config *dir_config = in_dir_config;
    char *content_type = apr_pstrdup( cmd->pool, type );

/* Probably don't need this bit... */
/*
    if (*ext == '.')
        ++ext;
*/

    ap_str_tolower(content_type);

    rec = find_extension( dir_config->extensions, ext );
    if( rec == NULL ) {
        rec = (extension_rec *) apr_array_push(dir_config->extensions);
        rec->extension = apr_pstrdup( cmd->pool, ext );
        rec->content_types = apr_array_make( cmd->pool, 10,
                                            sizeof( accept_rec ));
    }

    new = (accept_rec *) apr_array_push( rec->content_types );
    get_entry( cmd->pool, new, content_type );
    make_etag_hashcode( cmd->pool, new );

    return NULL;
}

/**
 * Set the minimun number of stars in an Accept token that should be
 * ignored when performing negotiation.
 *
 * @param in_dir_config the configuration information for this directory
 * @param stars the number of '*' tokens at which we start ignoring any
 * Accept tokens
 * @return NULL to indicate that processing was successful
 */

static const char *add_xhtml_ignore(cmd_parms *cmd, void *in_dir_config,
                                    const char *stars)
{
    xhtml_dir_config *dir_config = (xhtml_dir_config *)in_dir_config;
    int ignore_stars = 2;

    if( stars != NULL ) {
        ignore_stars = atoi(stars);
        if(( ignore_stars < 0 ) || ( ignore_stars > 3 )) {
            return apr_pstrcat(cmd->pool,
                    "XhtmlNegStarsIgnore ", stars,
                    " number of stars out of bounds", NULL);
        }
    }

    dir_config->stars_ignore = ignore_stars;
    return NULL;
}

/**
 * Set the log file name for this module.
 *
 * @param cmd the configuration information for this module on a per-server
 * basis
 * @param logfile the name of the log file to which the module should write
 * @return NULL to indicate that processing was successful
 */

static const char *add_xhtml_log(cmd_parms *cmd, void *dummy,
                                 const char *logfile)
{
    xhtml_neg_state *xns;

    xns = ap_get_module_config (cmd->server->module_config,
                                &xhtml_neg_module);

    xns->fname = apr_pstrdup( cmd->pool, logfile );
    xns->log_fd = NULL;

    return NULL;
}

/**
 * Create a default XHTML negotiation module configuration record.
 *
 * @param p a memory pool from which we can allocate memory that can later
 * be recycled
 * @return a new per-server configuration structure for this module
 */

static void *make_xhtml_neg_state (apr_pool_t *p, server_rec *s)
{
    xhtml_neg_state *xns =
      (xhtml_neg_state *)apr_palloc (p, sizeof (xhtml_neg_state));

    xns->fname = NULL;
    xns->log_fd = NULL;
    xns->cache = caching_dontcare;

    return (void *)xns;
}

/**
 * Create a default directory configuration module.
 *
 * @param p a memory pool from which we can allocate memory that can later
 * be recycled
 * @return a new per-directory configuration structure for this module
 */

static void *create_xhtml_dir_config(apr_pool_t *p, char *dummy)
{

    xhtml_dir_config *new =
        (xhtml_dir_config *) apr_pcalloc(p, sizeof(xhtml_dir_config));
    new->active = active_dontcare;
    new->extensions = apr_array_make(p, 5, sizeof(extension_rec));
    new->stars_ignore = -1;

    return (void *)new;
}

/**
 * Merge configuration info from different directories.
 *
 * @param p a memory pool from which we can allocate memory that can later
 * be recycled
 * @param basev the base directory configuration to be merged
 * @param addv the additional directory configuration to be merged
 * @return a merged per-directory configuration structure for this module
 */

static void *merge_xhtml_dir_configs(apr_pool_t *p, void *basev, void *addv)
{

    xhtml_dir_config *new = (xhtml_dir_config *) apr_pcalloc(p,
                                                 sizeof(xhtml_dir_config));
    xhtml_dir_config *base = (xhtml_dir_config *) basev;
    xhtml_dir_config *add = (xhtml_dir_config *) addv;
    extension_rec *item, *list;
    int i;

    if (add->active == active_dontcare) {
        new->active = base->active;
    } else {
        new->active = add->active;
    };

    new->extensions = apr_array_copy(p, add->extensions);

    if( add->stars_ignore >= 0 ) {
        new->stars_ignore = add->stars_ignore;
    } else {
        new->stars_ignore = base->stars_ignore;
    }

    /** Merge new file extensions into new array */
    list = (extension_rec*)base->extensions->elts;
    for (i = 0; i < base->extensions->nelts; ++i) {
        item = &list[i];
        if( find_extension( new->extensions, item->extension ) == NULL ) {
            extension_rec *new_item;
            new_item = (extension_rec *) apr_array_push(new->extensions);
            new_item->extension = item->extension;
            new_item->content_types = apr_array_copy(p, item->content_types);
        }
    }

    return new;
}

/**
 * Initialize the log file, if one is specified.
 *
 * @param s the server configuration from which we derive the state of this
 * module
 * @param p a memory pool from which we can allocate memory that can later
 * be recycled
 */

static int init_xhtml_log (apr_pool_t *pconf, apr_pool_t *plog,
                          apr_pool_t *ptemp, server_rec *s)
{
    apr_file_t *fd;
    char *fname;

    xhtml_neg_state *xns = ap_get_module_config (s->module_config,
                           &xhtml_neg_module);

    if (xns->log_fd != NULL) {
        return OK;
    }

    /* Don't open a log file if there's no filename. */
    if( mod_xhtml_strempty( xns->fname )) {
        return OK;
    }

    fname = ap_server_root_relative(pconf, xns->fname);
    fd = NULL;

    if( apr_file_open(&fd, fname, xfer_flags, xfer_perms, pconf)
                     != APR_SUCCESS) {
        ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
                     "could not open log file %s.", fname);
        exit(1);
    }

    xns->log_fd = fd;
    return OK;
}

/** Register specific hooks during request processing. */

static void xhtml_register_hooks(apr_pool_t *p)
{
    ap_hook_post_config(init_xhtml_log, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_handler(xhtml_negotiate, NULL, NULL, APR_HOOK_FIRST);
}

/** Record for registering commands with the Apache 2.0.x server. */

command_rec config_xhtml_cmds[] = {
    AP_INIT_FLAG("XhtmlNegActive", set_xhtml_active, NULL, OR_INDEXES,
                 "Is this module enabled? Defaults to no"),
    AP_INIT_TAKE1("XhtmlNegLog", add_xhtml_log, NULL, RSRC_CONF,
                  "A file name for an output log file"),
    AP_INIT_ITERATE2("XhtmlNegTypes", add_xhtml_type, NULL, OR_INDEXES,
                  "A file extension followed by one or more matching "
                  "content-type strings"),
    AP_INIT_TAKE1("XhtmlNegStarsIgnore", add_xhtml_ignore, NULL, OR_INDEXES,
                  "The number of stars in an Accept header which should be "
                  "ignored if we match them"),
    AP_INIT_FLAG("XhtmlNegCache", set_xhtml_cache_negotiated, NULL, RSRC_CONF,
                 "Should negotiated HTTP 1.0 requests be cacheable? "
                 "Defaults to no"),
    { NULL }
};

/** Module declaration for registering hooks into the Apache 2.0.x server. */

module AP_MODULE_DECLARE_DATA xhtml_neg_module =
{
    STANDARD20_MODULE_STUFF,
    create_xhtml_dir_config,    /**< dir config creater */
    merge_xhtml_dir_configs,    /**< dir config merger */
    make_xhtml_neg_state,       /**< server config */
    NULL,                       /**< merge server configs */
    config_xhtml_cmds,          /**< command table */
    xhtml_register_hooks,       /**< set up other request processing hooks */
};

